@@ -4,19 +4,21 @@ at some point in an evolutionary algorithm. You can apply
 evolutionary steps by directly calling methods on the population
 or by applying an `evol.Evolution` object.
 """
-from itertools import cycle, islice
-from math import ceil
-
 from abc import ABCMeta, abstractmethod
 from copy import copy
-from multiprocess.pool import Pool
+from itertools import cycle, islice
+from math import ceil
 from random import choices, randint
 from typing import Any, Callable, Generator, Iterable, Iterator, List, Optional, Sequence, TYPE_CHECKING
 from uuid import uuid4
 
+from multiprocess.pool import Pool
+
 from evol import Individual
+from evol.conditions import Condition
+from evol.exceptions import StopEvolution
+from evol.helpers.groups import group_random
 from evol.utils import offspring_generator, select_arguments
-from evol.logger import BaseLogger
 from evol.serialization import SimpleSerializer
 
 if TYPE_CHECKING:
@@ -31,7 +33,6 @@ class BasePopulation(metaclass=ABCMeta):
                  checkpoint_target: Optional[str] = None,
                  concurrent_workers: Optional[int] = 1,
                  maximize: bool = True,
-                 logger=None,
                  generation: int = 0,
                  intended_size: Optional[int] = None,
                  serializer=None):
@@ -42,7 +43,6 @@ class BasePopulation(metaclass=ABCMeta):
         self.id = str(uuid4())[:6]
         self.individuals = [Individual(chromosome=chromosome) for chromosome in chromosomes]
         self.intended_size = intended_size or len(self.individuals)
-        self.logger = logger or BaseLogger()
         self.maximize = maximize
         self.serializer = serializer or SimpleSerializer(target=checkpoint_target)
         self.pool = None if concurrent_workers == 1 else Pool(concurrent_workers)
@@ -76,6 +76,10 @@ class BasePopulation(metaclass=ABCMeta):
         for individual in self.individuals:
             yield individual.chromosome
 
+    @property
+    def is_evaluated(self) -> bool:
+        return all(individual.fitness is not None for individual in self)
+
     @classmethod
     def generate(cls,
                  init_function: Callable[[], Any],
@@ -142,9 +146,13 @@ class BasePopulation(metaclass=ABCMeta):
         :return: Population
         """
         result = copy(self)
-        for _ in range(n):
-            for step in evolution:
-                result = step.apply(result)
+        try:
+            for _ in range(n):
+                Condition.check(result)
+                for step in evolution:
+                    result = step.apply(result)
+        except StopEvolution:
+            pass
         return result
 
     @abstractmethod
@@ -183,7 +191,8 @@ class BasePopulation(metaclass=ABCMeta):
 
     def mutate(self,
                mutate_function: Callable[..., Any],
-               probability: float = 1.0, **kwargs) -> 'BasePopulation':
+               probability: float = 1.0,
+               elitist: bool = False, **kwargs) -> 'BasePopulation':
         """Mutate the chromosome of each individual.
 
         :param mutate_function: Function that accepts a chromosome and returns
@@ -191,23 +200,17 @@ class BasePopulation(metaclass=ABCMeta):
         :param probability: Probability that the individual mutates.
             The function is only applied in the given fraction of cases.
             Defaults to 1.0.
+        :param elitist: If True, do not mutate the current best individual(s).
+            Note that this only applies to evaluated individuals. Any unevaluated
+            individual will be treated as normal.
+            Defaults to False.
         :param kwargs: Arguments to pass to the mutation function.
         :return: self
         """
+        elite_fitness = self.current_best if elitist else None
         for individual in self.individuals:
-            individual.mutate(mutate_function, probability=probability, **kwargs)
-        return self
-
-    def log(self, **kwargs) -> 'BasePopulation':
-        """
-        Logs a population. If a Population object was initialized with a logger
-        object then you may specify how logging is handled. The base logging
-        operation just logs to standard out.
-
-        :return: self
-        """
-        self.evaluate(lazy=True)
-        self.logger.log(population=self, **kwargs)
+            if elite_fitness is None or individual.fitness != elite_fitness:
+                individual.mutate(mutate_function, probability=probability, **kwargs)
         return self
 
     def map(self, func: Callable[..., Individual], **kwargs) -> 'BasePopulation':
@@ -259,9 +262,10 @@ class BasePopulation(metaclass=ABCMeta):
             resulting_size = min(round(fraction * len(self.individuals)), n)
         self.evaluate(lazy=True)
         if resulting_size == 0:
-            raise RuntimeError('no one survived!')
+            raise RuntimeError(f'No individual out of {len(self.individuals)} survived!')
         if resulting_size > len(self.individuals):
-            raise ValueError('everyone survives! must provide "fraction" and/or "n" < population size')
+            raise ValueError(f'everyone survives in population {self.id}: '
+                             f'{resulting_size} out of {len(self.individuals)} must survive.')
         if luck:
             self.individuals = choices(self.individuals, k=resulting_size, weights=self._individual_weights)
         else:
@@ -269,7 +273,7 @@ class BasePopulation(metaclass=ABCMeta):
             self.individuals = sorted_individuals[:resulting_size]
         return self
 
-    def callback(self, callback_function: Callable[['BasePopulation'], None],
+    def callback(self, callback_function: Callable[..., None],
                  **kwargs) -> 'BasePopulation':
         """
         Performs a callback function on the population. Can be used for
@@ -279,9 +283,72 @@ class BasePopulation(metaclass=ABCMeta):
         :return:
         """
         self.evaluate(lazy=True)
-        callback_function(copy(self), **kwargs)
+        callback_function(self, **kwargs)
         return self
 
+    def group(self, grouping_function: Callable[..., List[List[int]]] = group_random,
+              **kwargs) -> List['BasePopulation']:
+        """
+        Group a population into islands.
+
+        Divides the population into multiple island populations, each of which
+        contains a subset of the original population. An individual from the
+        original population may end up in multiple (>= 0) island populations.
+
+        :param grouping_function: Function that allocates individuals to the
+            island populations. It will be passed a list of individuals plus
+            the kwargs passed to this method, and must return a list of lists
+            of integers, each sub-list representing an island and the integers
+            representing the index of an individual in the list. Each island
+            must contain at least one individual, and individual may be copied
+            to multiple islands.
+        :param kwargs: Additional keyworded arguments are passed to the
+            grouping function.
+        :return: List[Population]
+        """
+        group_indexes = grouping_function(self.individuals, **kwargs)
+        if len(group_indexes) == 0:
+            raise ValueError('Group yielded zero islands.')
+        result = [self._subset(index=index, subset_id=str(i)) for i, index in enumerate(group_indexes)]
+        return result
+
+    @classmethod
+    def combine(cls, *populations: 'BasePopulation',
+                intended_size: Optional[int] = None,
+                pool: Optional[Pool] = None) -> 'BasePopulation':
+        """
+        Combine multiple island populations into a single population.
+
+        The resulting population is reduced to its intended size.
+
+        :param populations: Populations to combine.
+        :param intended_size: Intended size of the resulting population.
+            Defaults to the sum of the intended sizes of the islands.
+        :param pool: Optionally provide a multiprocessing pool to be
+            used by the population.
+        :return: Population
+        """
+        if len(populations) == 0:
+            raise ValueError('Cannot combine zero islands into one.')
+        result = copy(populations[0])
+        for pop in populations[1:]:
+            result.individuals += pop.individuals
+        result.intended_size = intended_size or sum([pop.intended_size for pop in populations])
+        result.pool = pool
+        result.id = result.id.split('-')[0]
+        return result.survive(n=result.intended_size)
+
+    def _subset(self, index: List[int], subset_id: str) -> 'BasePopulation':
+        """Create a new population that is a subset of this population."""
+        if len(index) == 0:
+            raise ValueError('Grouping yielded an empty island.')
+        result = copy(self)
+        result.individuals = [result.individuals[i] for i in index]
+        result.intended_size = len(result.individuals)
+        result.pool = None  # Subsets shouldn't parallelize anything
+        result.id += '-' + subset_id
+        return result
+
     def _update_documented_best(self):
         """Update the documented best"""
         current_best = self.current_best
@@ -298,8 +365,6 @@ class Population(BasePopulation):
     :param eval_function: Function that reduces a chromosome to a fitness.
     :param maximize: If True, fitness will be maximized, otherwise minimized.
         Defaults to True.
-    :param logger: Logger object for the Population. If None, a new BaseLogger
-        is created. Defaults to None.
     :param generation: Generation of the Population. This is incremented after
         each breed call. Defaults to 0.
     :param intended_size: Intended size of the Population. The population will
@@ -317,7 +382,6 @@ class Population(BasePopulation):
                  chromosomes: Iterable,
                  eval_function: Callable[..., float],
                  maximize: bool = True,
-                 logger=None,
                  generation: int = 0,
                  intended_size: Optional[int] = None,
                  checkpoint_target: Optional[str] = None,
@@ -328,23 +392,23 @@ class Population(BasePopulation):
                          checkpoint_target=checkpoint_target,
                          concurrent_workers=concurrent_workers,
                          maximize=maximize,
-                         logger=logger,
                          generation=generation,
                          intended_size=intended_size,
                          serializer=serializer)
 
     def __copy__(self):
-        result = self.__class__(chromosomes=self.chromosomes,
+        result = self.__class__(chromosomes=[],
                                 eval_function=self.eval_function,
                                 maximize=self.maximize,
                                 serializer=self.serializer,
                                 intended_size=self.intended_size,
-                                logger=self.logger,
                                 generation=self.generation,
                                 concurrent_workers=1)  # Prevent new pool from being made
+        result.individuals = [copy(individual) for individual in self.individuals]
         result.concurrent_workers = self.concurrent_workers
         result.pool = self.pool
         result.documented_best = self.documented_best
+        result.id = self.id
         return result
 
     def evaluate(self, lazy: bool = False) -> 'Population':
@@ -370,15 +434,6 @@ class Population(BasePopulation):
         self._update_documented_best()
         return self
 
-    def apply(self, func: Callable[..., 'Population'], **kwargs) -> 'Population':
-        """Apply the provided function to the population.
-
-        :param func: A function to apply to the population, which returns the (modified) population.
-        :param kwargs: Arguments to pass to the function.
-        :return: self
-        """
-        return func(self, **kwargs)
-
 
 class Contest:
     """A single contest among a group of competitors.
@@ -451,8 +506,6 @@ class ContestPopulation(BasePopulation):
     :param contests_per_round: Minimum number of contests each individual
         takes part in for each evaluation round. The actual number of contests
         per round is a multiple of individuals_per_contest. Defaults to 10.
-    :param logger: Logger object for the Population. If None, a new BaseLogger
-        is created. Defaults to None.
     :param generation: Generation of the Population. This is incremented after
         echo survive call. Defaults to 0.
     :param intended_size: Intended size of the Population. The population will
@@ -465,15 +518,14 @@ class ContestPopulation(BasePopulation):
     :param concurrent_workers: If > 1, evaluate individuals in {concurrent_workers}
         separate processes. If None, concurrent_workers is set to n_cpus. Defaults to 1.
     """
-    eval_function: Callable[[Iterable[Any]], Sequence[float]]  # This population expects a different eval signature
+    eval_function: Callable[..., Sequence[float]]  # This population expects a different eval signature
 
     def __init__(self,
                  chromosomes: Iterable,
-                 eval_function: Callable[[Iterable[Any]], Sequence[float]],
+                 eval_function: Callable[..., Sequence[float]],
                  maximize: bool = True,
                  individuals_per_contest=2,
                  contests_per_round=10,
-                 logger=None,
                  generation: int = 0,
                  intended_size: Optional[int] = None,
                  checkpoint_target: Optional[int] = None,
@@ -482,7 +534,6 @@ class ContestPopulation(BasePopulation):
         super().__init__(chromosomes=chromosomes,
                          eval_function=eval_function,
                          maximize=maximize,
-                         logger=logger,
                          generation=generation,
                          intended_size=intended_size,
                          checkpoint_target=checkpoint_target,
@@ -492,19 +543,20 @@ class ContestPopulation(BasePopulation):
         self.individuals_per_contest = individuals_per_contest
 
     def __copy__(self):
-        result = self.__class__(chromosomes=self.chromosomes,
+        result = self.__class__(chromosomes=[],
                                 eval_function=self.eval_function,
                                 maximize=self.maximize,
                                 contests_per_round=self.contests_per_round,
                                 individuals_per_contest=self.individuals_per_contest,
                                 serializer=self.serializer,
                                 intended_size=self.intended_size,
-                                logger=self.logger,
                                 generation=self.generation,
                                 concurrent_workers=1)
+        result.individuals = [copy(individual) for individual in self.individuals]
         result.pool = self.pool
         result.concurrent_workers = self.concurrent_workers
         result.documented_best = None
+        result.id = self.id
         return result
 
     def evaluate(self,
